#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "cnheider"
__doc__ = ""


class IterItemsMixin(object):
    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value


class IterKeysMixin(object):
    def __iter__(self):
        for attr in self.__dict__.keys():
            yield attr


class IterValuesMixin(object):
    def __iter__(self):
        a = self.__dict__
        for value in self.__dict__.values():
            yield value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "cnheider"
__doc__ = ""
from warg import app_path

_app_name = "MyApp"
_app_author = __author__

props = ("user_data", "user_config", "user_cache", "user_state", "user_log", "site_data", "site_config")


def test_all():
    print("-- app dirs (with optional 'version')")
    dirs = app_path.AppPath(_app_name, _app_author, app_version="1.0", ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))


def test_no_ver():
    print("\n-- app dirs (without optional 'version')")
    dirs = app_path.AppPath(_app_name, _app_author, ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))


def test_author():
    print("\n-- app dirs (without optional '_app_author')")
    dirs = app_path.AppPath(_app_name, ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))


def test_no_author():
    print("\n-- app dirs (with disabled '_app_author')")
    dirs = app_path.AppPath(_app_name, app_author=False, ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))



from functools import partial, wraps
from inspect import signature
from typing import Callable, Iterable, Mapping

from sorcery import dict_of

from warg.named_ordered_dictionary import NOD


def dict_of_func(func: callable = None) -> callable:
    if not func:
        new_debug_func = partial(dict_of_func)
        return new_debug_func

    @wraps(func)
    def wrapper(*args, **kwargs):
        if func.__name__ is not "__init__":
            return_value = func(*args, **kwargs)
            if isinstance(return_value, NOD):
                return return_value
            elif isinstance(return_value, Mapping):
                return NOD(**return_value)
            elif isinstance(return_value, Iterable):
                return NOD(*return_value)
            return NOD(return_value)
        else:
            return func(*args, **kwargs)

    sig = signature(func)
    sig = sig.replace(return_annotation=NOD)
    func.__signature__ = sig
    wrapper.__signature__ = sig

    return wrapper


def nod_return_class(cls):
    for [method_key, method_value] in vars(cls).items():
        if isinstance(method_value, Callable):
            setattr(cls, method_key, dict_of_func(method_value))
    return cls


class MetaNodReturnBase(type):
    def __new__(cls, *args, **kwargs):
        class_obj = super().__new__(cls, *args, **kwargs)
        class_obj = nod_return_class(class_obj)
        return class_obj


class AutoNodReturns(metaclass=MetaNodReturnBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


if __name__ == "__main__":

    class NodReturnExampleClass(AutoNodReturns):
        def cool(self, agkjas):
            slfs = 4 + agkjas
            return NOD(4, **dict_of(slfs))

        def hot(self):
            s = 7
            a = 2
            return dict_of(s, a)

        def luke_warm(self, a, b):
            return 1, 3, 5

    auto_nod_return = NodReturnExampleClass()

    yes = auto_nod_return.cool(5)
    print(yes)
    assert yes.arg0 == 4
    assert yes.arg0 == yes.as_list()[0]

    no = auto_nod_return.hot()
    print(no)
    assert no.a == 2
    assert no.s == 7

    maybe = auto_nod_return.luke_warm(2, 8)
    print(maybe)
    assert maybe.arg1 == 3
    assert maybe.arg2 == 5



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pathlib

__author__ = "cnheider"
__doc__ = "Application data directories extension for pathlib"

import os
from warg.app_path_utilities import get_win_folder, SYSTEM


class AppPath(object):
    r"""
AppPath class for easing cross platform access to proper app data directories
"""

    def __init__(
        self,
        app_name=None,
        app_author=None,
        app_version=None,
        roaming=False,
        multi_path=False,
        ensure_existence=True,
    ):
        self._app_name = app_name
        self._app_author = app_author
        self._app_version = app_version
        self._roaming = roaming
        self._multi_path = multi_path
        self._ensure_existence = ensure_existence

    @staticmethod
    def ensure_existence(enabled, out):
        if enabled:
            if not out.exists():
                out.mkdir(parents=True)

    @property
    def user_data(self) -> pathlib.Path:
        out = self.user_data_path(
            self._app_name, self._app_author, version=self._app_version, roaming=self._roaming
        )

        self.ensure_existence(self._ensure_existence, out)
        return out

    @property
    def site_data(self) -> pathlib.Path:
        out = self.site_data_path(
            self._app_name, self._app_author, version=self._app_version, multi_path=self._multi_path
        )

        self.ensure_existence(self._ensure_existence, out)

        return out

    @property
    def user_config(self) -> pathlib.Path:
        out = self.user_config_path(
            self._app_name, self._app_author, version=self._app_version, roaming=self._roaming
        )
        self.ensure_existence(self._ensure_existence, out)
        return out

    @property
    def site_config(self) -> pathlib.Path:
        out = self.site_config_path(
            self._app_name, self._app_author, version=self._app_version, multi_path=self._multi_path
        )
        self.ensure_existence(self._ensure_existence, out)
        return out

    @property
    def user_cache(self) -> pathlib.Path:
        out = self.user_cache_path(self._app_name, self._app_author, version=self._app_version)
        self.ensure_existence(self._ensure_existence, out)
        return out

    @property
    def user_state(self) -> pathlib.Path:
        out = self.user_state_path(self._app_name, self._app_author, version=self._app_version)
        self.ensure_existence(self._ensure_existence, out)
        return out

    @property
    def user_log(self) -> pathlib.Path:
        out = self.user_log_path(self._app_name, self._app_author, version=self._app_version)
        self.ensure_existence(self._ensure_existence, out)
        return out

    @staticmethod
    def user_data_path(app_name=None, app_author=None, version=None, roaming=False) -> pathlib.Path:
        r"""Return full path to the user-specific data dir for this application.

    "app_name" is the name of application.
        If None, just the system directory is returned.
    "app_author" (only used on Windows) is the name of the
        app_author or distributing body for this application. Typically
        it is the owning company name. This falls back to app_name. You may
        pass False to disable it.
    "version" is an optional version path element to append to the
        path. You might want to use this if you want multiple versions
        of your app to be able to run independently. If used, this
        would typically be "<major>.<minor>".
        Only applied when app_name is present.
    "roaming" (boolean, default False) can be set True to use the Windows
        roaming appdata directory. That means that for users on a Windows
        network setup for roaming profiles, this user data will be
        sync'd on login. See
        <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx>
        for a discussion of issues.


  Notes:
  - MSDN on where to store app data files:
  http://support.microsoft.com/default.aspx?scid=kb;en-us;310294#XSLTH3194121123120121120120
  - Mac OS X: http://developer.apple.com/documentation/MacOSX/Conceptual/BPFileSystem/index.html
  - XDG spec for Un*x: https://standards.freedesktop.org/basedir-spec/basedir-spec-latest.html

Typical user data directories are:
    Mac OS X:               ~/Library/Application Support/<AppName>
    Unix:                   ~/.local/share/<AppName>    # or in $XDG_DATA_HOME, if defined
    Win XP (not roaming):   C:\Documents and Settings\<username>\Application Data\<AppAuthor>\<AppName>
    Win XP (roaming):       C:\Documents and Settings\<username>\Local Settings\Application
    Data\<AppAuthor>\<AppName>
    Win 7  (not roaming):   C:\Users\<username>\AppData\Local\<AppAuthor>\<AppName>
    Win 7  (roaming):       C:\Users\<username>\AppData\Roaming\<AppAuthor>\<AppName>

For Unix, we follow the XDG spec and support $XDG_DATA_HOME.
That means, by default "~/.local/share/<AppName>".
"""

        if SYSTEM == "win32":
            if app_author is None:
                app_author = app_name
            const = roaming and "CSIDL_APPDATA" or "CSIDL_LOCAL_APPDATA"
            path = pathlib.Path(os.path.normpath(get_win_folder(const)))
            if app_name:
                if app_author is not False:
                    path = path / app_author / app_name
                else:
                    path = path / app_name
        elif SYSTEM == "darwin":
            path = pathlib.Path(os.path.expanduser("~/Library/Application Support/"))
            if app_name:
                path = path / app_name
        else:
            path = pathlib.Path(os.getenv("XDG_DATA_HOME", os.path.expanduser("~/.local/share")))
            if app_name:
                path = path / app_name
        if app_name and version:
            path = path / version
        return path

    @staticmethod
    def site_data_path(app_name=None, app_author=None, version=None, multi_path=False) -> pathlib.Path:
        r"""Return full path to the user-shared data dir for this application.

    "app_name" is the name of application.
        If None, just the system directory is returned.
    "app_author" (only used on Windows) is the name of the
        app_author or distributing body for this application. Typically
        it is the owning company name. This falls back to app_name. You may
        pass False to disable it.
    "version" is an optional version path element to append to the
        path. You might want to use this if you want multiple versions
        of your app to be able to run independently. If used, this
        would typically be "<major>.<minor>".
        Only applied when app_name is present.
    "multi_path" is an optional parameter only applicable to *nix
        which indicates that the entire list of data dirs should be
        returned. By default, the first item from XDG_DATA_DIRS is
        returned, or '/usr/local/share/<AppName>',
        if XDG_DATA_DIRS is not set

Typical site data directories are:
    Mac OS X:   /Library/Application Support/<AppName>
    Unix:       /usr/local/share/<AppName> or /usr/share/<AppName>
    Win XP:     C:\Documents and Settings\All Users\Application Data\<AppAuthor>\<AppName>
    Vista:      (Fail! "C:\ProgramData" is a hidden *system* directory on Vista.)
    Win 7:      C:\ProgramData\<AppAuthor>\<AppName>   # Hidden, but writeable on Win 7.

For Unix, this is using the $XDG_DATA_DIRS[0] default.

WARNING: Do not use this on Windows. See the Vista-Fail note above for why.
"""
        if SYSTEM == "win32":
            if app_author is None:
                app_author = app_name
            path = pathlib.Path(os.path.normpath(get_win_folder("CSIDL_COMMON_APPDATA")))
            if app_name:
                if app_author is not False:
                    path = path / app_author / app_name
                else:
                    path = path / app_name
        elif SYSTEM == "darwin":
            path = pathlib.Path(os.path.expanduser("/Library/Application Support"))
            if app_name:
                path = path / app_name
        else:
            # XDG default for $XDG_DATA_DIRS
            # only first, if multipath is False
            path = os.getenv("XDG_DATA_DIRS", os.pathsep.join(["/usr/local/share", "/usr/share"]))
            path_list = [os.path.expanduser(x.rstrip(os.sep)) for x in path.split(os.pathsep)]
            if app_name:
                if version:
                    app_name = os.path.join(app_name, version)
                path_list = [os.sep.join([x, app_name]) for x in path_list]

            path_list = [pathlib.Path(a) for a in path_list]

            if multi_path:
                path = os.pathsep.join(path_list)
            else:
                path = path_list[0]
            return path

        if app_name and version:
            path = path / version
        return path

    @staticmethod
    def user_config_path(app_name=None, app_author=None, version=None, roaming=False) -> pathlib.Path:
        r"""Return full path to the user-specific config dir for this application.

    "app_name" is the name of application.
        If None, just the system directory is returned.
    "app_author" (only used on Windows) is the name of the
        app_author or distributing body for this application. Typically
        it is the owning company name. This falls back to app_name. You may
        pass False to disable it.
    "version" is an optional version path element to append to the
        path. You might want to use this if you want multiple versions
        of your app to be able to run independently. If used, this
        would typically be "<major>.<minor>".
        Only applied when app_name is present.
    "roaming" (boolean, default False) can be set True to use the Windows
        roaming appdata directory. That means that for users on a Windows
        network setup for roaming profiles, this user data will be
        sync'd on login. See
        <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx>
        for a discussion of issues.

Typical user config directories are:
    Mac OS X:               ~/Library/Preferences/<AppName>
    Unix:                   ~/.config/<AppName>     # or in $XDG_CONFIG_HOME, if defined
    Win *:                  same as user_data_dir

For Unix, we follow the XDG spec and support $XDG_CONFIG_HOME.
That means, by default "~/.config/<AppName>".
"""
        if SYSTEM == "win32":
            path = AppPath.user_data_path(app_name, app_author, None, roaming)
        elif SYSTEM == "darwin":
            path = pathlib.Path(os.path.expanduser("~/Library/Preferences/"))
            if app_name:
                path = path / app_name
        else:
            path = pathlib.Path(os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config")))
            if app_name:
                path = path / app_name
        if app_name and version:
            path = path / version
        return path

    @staticmethod
    def site_config_path(app_name=None, app_author=None, version=None, multi_path=False) -> pathlib.Path:
        r"""Return full path to the user-shared data dir for this application.

    "app_name" is the name of application.
        If None, just the system directory is returned.
    "app_author" (only used on Windows) is the name of the
        app_author or distributing body for this application. Typically
        it is the owning company name. This falls back to appname. You may
        pass False to disable it.
    "version" is an optional version path element to append to the
        path. You might want to use this if you want multiple versions
        of your app to be able to run independently. If used, this
        would typically be "<major>.<minor>".
        Only applied when app_name is present.
    "multi_path" is an optional parameter only applicable to *nix
        which indicates that the entire list of config dirs should be
        returned. By default, the first item from XDG_CONFIG_DIRS is
        returned, or '/etc/xdg/<AppName>', if XDG_CONFIG_DIRS is not set

Typical site config directories are:
    Mac OS X:   same as site_data_dir
    Unix:       /etc/xdg/<AppName> or $XDG_CONFIG_DIRS[i]/<AppName> for each value in
                $XDG_CONFIG_DIRS
    Win *:      same as site_data_dir
    Vista:      (Fail! "C:\ProgramData" is a hidden *system* directory on Vista.)

For Unix, this is using the $XDG_CONFIG_DIRS[0] default, if multipath=False

WARNING: Do not use this on Windows. See the Vista-Fail note above for why.
"""
        if SYSTEM == "win32":
            path = AppPath.site_data_path(app_name, app_author)
            if app_name and version:
                path = path / version
        elif SYSTEM == "darwin":
            path = pathlib.Path(os.path.expanduser("/Library/Preferences"))
            if app_name:
                path = path / app_name
        else:
            # XDG default for $XDG_CONFIG_DIRS
            # only first, if multi_path is False
            path = os.getenv("XDG_CONFIG_DIRS", "/etc/xdg")
            path_list = [os.path.expanduser(x.rstrip(os.sep)) for x in path.split(os.pathsep)]
            if app_name:
                if version:
                    app_name = os.path.join(app_name, version)
                path_list = [os.sep.join([x, app_name]) for x in path_list]

            path_list = [pathlib.Path(a) for a in path_list]

            if multi_path:
                path = os.pathsep.join(path_list)
            else:
                path = path_list[0]
        return path

    @staticmethod
    def user_cache_path(app_name=None, app_author=None, version=None, opinion=True) -> pathlib.Path:
        r"""Return full path to the user-specific cache dir for this application.

    "appname" is the name of application.
        If None, just the system directory is returned.
    "appauthor" (only used on Windows) is the name of the
        appauthor or distributing body for this application. Typically
        it is the owning company name. This falls back to appname. You may
        pass False to disable it.
    "version" is an optional version path element to append to the
        path. You might want to use this if you want multiple versions
        of your app to be able to run independently. If used, this
        would typically be "<major>.<minor>".
        Only applied when appname is present.
    "opinion" (boolean) can be False to disable the appending of
        "Cache" to the base app data dir for Windows. See
        discussion below.

Typical user cache directories are:
    Mac OS X:   ~/Library/Caches/<AppName>
    Unix:       ~/.cache/<AppName> (XDG default)
    Win XP:     C:\Documents and Settings\<username>\Local Settings\Application
    Data\<AppAuthor>\<AppName>\Cache
    Vista:      C:\Users\<username>\AppData\Local\<AppAuthor>\<AppName>\Cache

On Windows the only suggestion in the MSDN docs is that local settings go in
the `CSIDL_LOCAL_APPDATA` directory. This is identical to the non-roaming
app data dir (the default returned by `user_data_dir` above). Apps typically
put cache data somewhere *under* the given dir here. Some examples:
    ...\Mozilla\Firefox\Profiles\<ProfileName>\Cache
    ...\Acme\SuperApp\Cache\1.0
OPINION: This function appends "Cache" to the `CSIDL_LOCAL_APPDATA` value.
This can be disabled with the `opinion=False` option.
"""
        if SYSTEM == "win32":
            if app_author is None:
                app_author = app_name
            path = pathlib.Path(os.path.normpath(get_win_folder("CSIDL_LOCAL_APPDATA")))
            if app_name:
                if app_author is not False:
                    path = path / app_author / app_name
                else:
                    path = path / app_name
                if opinion:
                    path = path / "Cache"
        elif SYSTEM == "darwin":
            path = pathlib.Path(os.path.expanduser("~/Library/Caches"))
            if app_name:
                path = path / app_name
        else:
            path = pathlib.Path(os.getenv("XDG_CACHE_HOME", os.path.expanduser("~/.cache")))
            if app_name:
                path = path / app_name
        if app_name and version:
            path = path / version
        return path

    @staticmethod
    def user_state_path(app_name=None, app_author=None, version=None, roaming=False) -> pathlib.Path:
        r"""Return full path to the user-specific state dir for this application.

    "app_name" is the name of application.
        If None, just the system directory is returned.
    "app_author" (only used on Windows) is the name of the
        app_author or distributing body for this application. Typically
        it is the owning company name. This falls back to appname. You may
        pass False to disable it.
    "version" is an optional version path element to append to the
        path. You might want to use this if you want multiple versions
        of your app to be able to run independently. If used, this
        would typically be "<major>.<minor>".
        Only applied when app_name is present.
    "roaming" (boolean, default False) can be set True to use the Windows
        roaming appdata directory. That means that for users on a Windows
        network setup for roaming profiles, this user data will be
        sync'd on login. See
        <http://technet.microsoft.com/en-us/library/cc766489(WS.10).aspx>
        for a discussion of issues.

Typical user state directories are:
    Mac OS X:  same as user_data_dir
    Unix:      ~/.local/state/<AppName>   # or in $XDG_STATE_HOME, if defined
    Win *:     same as user_data_dir

For Unix, we follow this Debian proposal <https://wiki.debian.org/XDGBaseDirectorySpecification#state>
to extend the XDG spec and support $XDG_STATE_HOME.

That means, by default "~/.local/state/<AppName>".
"""
        if SYSTEM in ["win32", "darwin"]:
            path = AppPath.user_data_path(app_name, app_author, None, roaming)
        else:
            path = pathlib.Path(os.getenv("XDG_STATE_HOME", os.path.expanduser("~/.local/state")))
            if app_name:
                path = path / app_name
        if app_name and version:
            path = path / version
        return path

    @staticmethod
    def user_log_path(app_name=None, app_author=None, version=None, opinion=True) -> pathlib.Path:
        r"""Return full path to the user-specific log dir for this application.

    "app_name" is the name of application.
        If None, just the system directory is returned.
    "app_author" (only used on Windows) is the name of the
        app_author or distributing body for this application. Typically
        it is the owning company name. This falls back to appname. You may
        pass False to disable it.
    "version" is an optional version path element to append to the
        path. You might want to use this if you want multiple versions
        of your app to be able to run independently. If used, this
        would typically be "<major>.<minor>".
        Only applied when app_name is present.
    "opinion" (boolean) can be False to disable the appending of
        "Logs" to the base app data dir for Windows, and "log" to the
        base cache dir for Unix. See discussion below.

Typical user log directories are:
    Mac OS X:   ~/Library/Logs/<AppName>
    Unix:       ~/.cache/<AppName>/log  # or under $XDG_CACHE_HOME if defined
    Win XP:     C:\Documents and Settings\<username>\Local Settings\Application
    Data\<AppAuthor>\<AppName>\Logs
    Vista:      C:\Users\<username>\AppData\Local\<AppAuthor>\<AppName>\Logs

On Windows the only suggestion in the MSDN docs is that local settings
go in the `CSIDL_LOCAL_APPDATA` directory. (Note: I'm interested in
examples of what some windows apps use for a logs dir.)

OPINION: This function appends "Logs" to the `CSIDL_LOCAL_APPDATA`
value for Windows and appends "log" to the user cache dir for Unix.
This can be disabled with the `opinion=False` option.
"""
        if SYSTEM == "darwin":
            path = pathlib.Path.joinpath(pathlib.Path(os.path.expanduser("~/Library/Logs")), app_name)
        elif SYSTEM == "win32":
            path = AppPath.user_data_path(app_name, app_author, version)
            version = False
            if opinion:
                path = path / "Logs"
        else:
            path = AppPath.user_cache_path(app_name, app_author, version)
            version = False
            if opinion:
                path = path / "log"
        if app_name and version:
            path = path / version
        return path


if __name__ == "__main__":
    _app_name = "MyApp"
    _app_author = __author__

    props = ("user_data", "user_config", "user_cache", "user_state", "user_log", "site_data", "site_config")

    print("-- app dirs (with optional 'version')")
    dirs = AppPath(_app_name, _app_author, app_version="1.0", ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))

    print("\n-- app dirs (without optional 'version')")
    dirs = AppPath(_app_name, _app_author, ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))

    print("\n-- app dirs (without optional '_app_author')")
    dirs = AppPath(_app_name, ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))

    print("\n-- app dirs (with disabled '_app_author')")
    dirs = AppPath(_app_name, app_author=False, ensure_existence=False)
    for prop in props:
        print("%s: %s" % (prop, getattr(dirs, prop)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "cnheider"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

import unittest

import warg.toaster as toaster


class ModuleTests(unittest.TestCase):
    """Test module level functions.
  """

    def setUp(self):
        toaster.init("toaster test suite")

    def test_init_uninit(self):
        assert toaster.is_initted()
        self.assertEqual(toaster.get_app_name(), "toaster test suite")

        toaster.de_init()
        assert not toaster.is_initted()

    def test_get_server_info(self):
        r = toaster.get_server_info()
        assert isinstance(r, dict), type(r)

    def test_get_server_caps(self):
        r = toaster.get_server_caps()
        assert isinstance(r, list), type(r)


class NotificationTests(unittest.TestCase):
    """Test notifications.
  """

    def setUp(self):
        toaster.init("toaster test suite")

    def test_basic(self):
        n = toaster.Toast("Title", "Body")
        n.show()
        n.close()

    def test_icon(self):
        n = toaster.Toast("Title", "Body", icon="notification-message-im")
        n.show()
        n.close()

    def test_icon_only(self):
        if "x-canonical-private-icon-only" in toaster.get_server_caps():
            n = toaster.Toast("", "", icon="notification-device-eject")
            n.set_hint_string("x-canonical-private-icon-only", "true")
            n.show()

    def test_urgency(self):
        nl = toaster.Toast("Low", "")
        nl.set_urgency(toaster.URGENCY_LOW)
        nl.show()
        nl.close()

        nn = toaster.Toast("Normal", "")
        nn.set_urgency(toaster.URGENCY_NORMAL)
        nn.show()
        nn.close()

        nu = toaster.Toast("Urgent", "")
        nu.set_urgency(toaster.URGENCY_CRITICAL)
        nu.show()
        nu.close()

    def test_update(self):
        n = toaster.Toast("First message", "Some text", icon="notification-message-im")
        n.show()

        # The icon should stay the same with this
        n.update("Second message", "Some more text")
        n.show()

        # But this should replace the icon
        n.update("Third message", "Yet more text, new icon.", icon="notification-message-email")
        n.show()
        n.close()

    def test_category(self):
        n = toaster.Toast("Plain")
        n.set_category("im.received")
        n.show()
        n.close()

    def test_timeout(self):
        n = toaster.Toast("Plain")
        self.assertEqual(n.get_timeout(), toaster.EXPIRES_DEFAULT)
        n.set_timeout(toaster.EXPIRES_NEVER)
        self.assertEqual(n.get_timeout(), toaster.EXPIRES_NEVER)
        n.set_timeout(5000)
        self.assertEqual(n.get_timeout(), 5000)
        n.show()
        n.close()

    def test_data(self):
        n = toaster.Toast("Plain")
        n._data["a"] = 1
        n.set_data("b", 2)  # pynotify API
        n.show()
        self.assertEqual(n.get_data("a"), 1)  # pynotify API
        self.assertEqual(n._data["b"], 2)
        n.close()

    def test_icon_from_pixbuf(self):
        import gi

        gi.require_version("Gtk", "3.0")
        from gi.repository import Gtk

        helper = Gtk.Button()
        pb = helper.render_icon(Gtk.STOCK_DIALOG_INFO, Gtk.IconSize.DIALOG)

        # pb = GdkPixbuf.Pixbuf.new_from_file("examples/applet-critical.png")
        n = toaster.Toast("Icon", "Testing icon from pixbuf")
        n.set_icon_from_pixbuf(pb)
        n.show()
        n.close()

    def test_set_location(self):
        n = toaster.Toast("Location", "Test setting location")
        n.set_location(320, 240)
        n.show()
        n.close()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


__author__ = "cnheider"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""


def action_callback(nid, action, notifications_registry):
    nid, action = int(nid), str(action)
    try:
        n = notifications_registry[nid]
    except KeyError:
        # this message was created through some other program.
        return
    n.action_callback(action, notifications_registry)


def closed_callback(nid, reason, notifications_registry):
    nid, reason = int(nid), int(reason)
    try:
        n = notifications_registry[nid]
    except KeyError:
        # this message was created through some other program.
        return
    n.closed_callback(n)
    del notifications_registry[nid]


def no_op(*args):
    """No-op function for callbacks.
  """
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from collections import Mapping
from typing import Any, Iterable, TypeVar

import sorcery

__author__ = "cnheider"

LOCALS = (
    "as_list",
    "as_dict",
    "as_tuples",
    "as_flat_tuples",
    "add_unnamed_arg",
    "dict_of",
    "keys",
    "update",
    "__setattr__",
)


class IllegalAttributeKey(Exception):
    def __init__(self, key, type):
        msg = f'Overwritting of attribute "{key}" on type "{type}" is not allowed'
        Exception.__init__(self, msg)


T = TypeVar("T", bound="NamedOrderedDictionary")


class NamedOrderedDictionary(Mapping):
    """

Usage:

.. code-block:: python

nodict = NamedOrderedDictionary()
nodict.paramA = 'str_parameter'
nodict.paramB = 10
assert nodict.paramA == 'str_parameter'
assert nodict.paramB == 10


.. code-block:: python

nodict = NamedOrderedDictionary()
nodict['paramA'] = 10
assert nodict.paramA == 10

.. code-block:: python

nodict = NamedOrderedDictionary({'paramA': 'str_parameter', 'paramB': 10})
assert nodict.paramA == 'str_parameter'
assert nodict.paramB == 10

.. code-block:: python

nodict = NamedOrderedDictionary(paramA='str_parameter', paramB=10)
assert nodict.paramA == 'str_parameter'
assert nodict.paramB == 10

.. code-block:: python

nodict = NamedOrderedDictionary('str_parameter', 10)
assert nodict.arg0 == 'str_parameter'
assert nodict.arg1 == 10

.. code-block:: python

arg0,arg1 = NamedOrderedDictionary('str_parameter', 10).as_list()
assert arg0 == 'str_parameter'
assert arg1 == 10

As with dictionaries you can use the `update()` method.

.. code-block:: python
nodict = NamedOrderedDictionary()
nodict.update({'paramA': 20, 'paramB': 'other_param', 'paramC': 5.0})
assert nodict.paramA == 20
assert nodict.paramB == 'other_param'

.. code-block:: python

nodict = NamedOrderedDictionary('str_parameter', 10)
nodict.update({'arg1': 20, 'arg0': 'other_param'})
assert nodict.arg0 == 'other_param'
assert nodict.arg1 == 20

.. code-block:: python

nodict = NamedOrderedDictionary(paramA='str_parameter', paramB=10)
nodict.update(20,'other_param')
assert nodict.paramB == 'other_param'
assert nodict.paramA == 20

"""

    # __slots__ = ('_unnamed_arg_i','__dict__')

    # _unnamed_arg_i = 0

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        # super().__init__(**kwargs)

        if len(args) == 1 and isinstance(args[0], dict):
            args_dict = args[0]
        else:
            args_dict = {}
            if len(args) == 1 and isinstance(args[0], Iterable):
                args = args[0]
            for arg in args:
                args_dict[f"arg{id(arg)}"] = arg

        args_dict.update(kwargs)
        self.update(args_dict or {})

    def as_list(self):
        return list(self.__dict__.values())

    def as_dict(self):
        return self.__dict__

    def as_tuples(self):
        return [(k, v) for (k, v) in self.__dict__.items()]

    def as_flat_tuples(self):
        return [(k, *v) for (k, v) in self.__dict__.items()]

    def add_unnamed_arg(self, arg):
        self.__dict__[f"arg{id(arg)}"] = arg

    @sorcery.spell
    def nod_of(frame_info, *args, **kwargs) -> T:
        """
Instead of:

{'foo': foo, 'bar': bar, 'spam': thing()}

or:

dict(foo=foo, bar=bar, spam=thing())

write:

NOD.dict_of(foo, bar, spam=thing())

In other words, returns a NamedOrderedDictionary with an item for each argument,
where positional arguments use their names as keys,
and keyword arguments do the same as in the usual dict constructor.

The positional arguments can be any of:

- plain variables,
- attributes, or
- subscripts (square bracket access) with string literal keys

So the following:

NOD.dict_of(spam, x.foo, y['bar'])

is equivalent to:

NOD.dict_of(spam=spam, foo=x.foo, bar=y['bar'])

*args are not allowed.

:rtype: object

"""

        result = sorcery.dict_of.at(frame_info)(*args, **kwargs)
        return NamedOrderedDictionary(result)

    def __getattr__(self, item):
        return self.__dict__[item]

    def __len__(self):
        return len(self.__dict__)

    def __setattr__(self, key, value):
        if key in LOCALS:
            raise IllegalAttributeKey(key, type=NamedOrderedDictionary.__name__)

        self.__dict__[key] = value

    def __getitem__(self, item) -> Any:
        return self.__dict__[item]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def keys(self):
        return self.__dict__.keys()

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        for key, value in self.__dict__.items():
            yield key, value

    def __repr__(self):
        print_str = f"{self.__class__.__name__}("
        for key, value in self:
            print_str += f"'{key}': {value}, "
        print_str = print_str[:-2]
        print_str += ")"
        return print_str

    def update(self, *args: Any, **kwargs: Any) -> None:
        """
Merge two attributes, overriding any repeated keys from
the `items` parameter.

Args:
items (dict): Python dictionary containing updated values.
"""

        if len(args) == 1 and isinstance(args[0], dict):
            args_dict = args[0]
        else:
            args_dict = {}
            a = list(self.__dict__.keys())
            for arg, key in zip(args, a):
                args_dict[key] = arg

        args_dict.update(kwargs)

        for key in args_dict:
            if key in LOCALS:
                raise IllegalAttributeKey(key, type=NamedOrderedDictionary.__name__)

        self.__dict__.update(args_dict)

    def __add__(self, other):
        if isinstance(other, NamedOrderedDictionary):
            for k in other.keys():
                if k in self.__dict__:
                    self.__dict__[k] += other.__dict__[k]
                else:
                    self.__dict__[k] = other.__dict__[k]
        elif isinstance(other, Iterable):
            for arg in other:
                self.add_unnamed_arg(arg)
        else:
            self.add_unnamed_arg(other)
        return self.__dict__

    def __sub__(self, other):
        if isinstance(other, NamedOrderedDictionary):
            for k in other.keys():
                if k in self.__dict__:
                    self.__dict__[k] -= other.__dict__[k]
        else:
            raise ArithmeticError(f"Can not subtract {type(other)} from {type(self)}")

    def __truediv__(self, other):
        if isinstance(other, str):
            return self.get(other)
        else:
            raise ArithmeticError(f"Can not access with {type(other)} in {type(self)}")

    def __floordiv__(self, other):
        return self.__truediv__(other)


NOD = NamedOrderedDictionary

if __name__ == "__main__":

    nodict = NamedOrderedDictionary()
    nodict.paramA = "str_parameter"
    nodict.paramB = 10
    assert nodict.paramA == "str_parameter"
    assert nodict.paramB == 10

    nodict = NamedOrderedDictionary({"paramA": "str_parameter", "paramB": 10})
    assert nodict.paramA == "str_parameter"
    assert nodict.paramB == 10

    nodict = NamedOrderedDictionary()
    nodict.update({"paramA": 20, "paramB": "other_param", "paramC": 5.0})
    nodict.paramA = 10
    assert nodict.paramA == 10
    assert nodict.paramB == "other_param"

    nodict = NamedOrderedDictionary()
    nodict["paramA"] = 10
    assert nodict.paramA == 10

    values = [1, 3, 5]
    nodict = NamedOrderedDictionary(10, val2=2, *values)
    nolist = nodict.as_list()
    assert nolist[0] == 10
    assert nodict.val2 == 2
    assert nolist[-2] == values[-1]

    nodict = NamedOrderedDictionary("str_parameter", 10)
    odict = {**nodict}
    assert nodict.as_list()[0] == list(odict.values())[0]

    nodict = NamedOrderedDictionary("str_parameter", 10)
    nodict.update(arg1=20, arg0="other_param")
    assert nodict.arg0 == "other_param"
    assert nodict.arg1 == 20

    nodict = NamedOrderedDictionary("str_parameter", 10)
    nodict.update({"arg1": 20, "arg0": "other_param"})
    assert nodict.arg0 == "other_param"
    assert nodict.arg1 == 20

    nodict = NamedOrderedDictionary(paramA="str_parameter", paramB=10)
    nodict.update(20, "other_param")
    assert nodict.paramB == "other_param"
    assert nodict.paramA == 20
    assert nodict.get("paramC") is None

    nodict = NamedOrderedDictionary(paramC="str_parameter", **nodict)
    assert nodict.paramB == "other_param"
    assert nodict.paramA == 20
    assert nodict.get("paramC") is not None

    arg0, arg1 = NamedOrderedDictionary("str_parameter", 10).as_list()
    assert arg0 == "str_parameter"
    assert arg1 == 10

    columns = NamedOrderedDictionary.nod_of(arg1, aræa=arg0)
    assert columns["arg1"] == arg1
    assert columns.arg1 == arg1
    assert columns["aræa"] == arg0
    assert columns / "aræa" == arg0
    assert id(columns / "aræa") == id(columns["aræa"])

    a = NamedOrderedDictionary(4, 2)
    a.list = "a"
    assert a.list == "a"

    a = NamedOrderedDictionary(4, 2)
    b = a + a
    assert a.as_list() == [8, 4]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import argparse
from pathlib import Path, PosixPath
from warnings import warn

import wrapt

__author__ = "cnheider"


class UpperAttrMetaclass(type):
    def __new__(cls, clsname, bases, dct):

        uppercase_attr = {}
        for name, val in dct.items():
            if not name.startswith("__"):
                uppercase_attr[name.upper()] = val
            else:
                uppercase_attr[name] = val

        return super(UpperAttrMetaclass, cls).__new__(cls, clsname, bases, uppercase_attr)


class ConfigObject(object):
    pass


def to_lower_properties(C_dict):
    if not isinstance(C_dict, dict):
        C_dict = to_dict(C_dict)

    a = ConfigObject()

    for (k, v) in C_dict.items():
        assert isinstance(k, str)
        lowered = k.lower()
        if isinstance(v, (PosixPath, Path)):
            setattr(a, lowered, str(v))
        else:
            setattr(a, lowered, v)

    return a


def get_upper_case_vars_or_protected_of(module):
    v = vars(module)
    if v:
        return {
            key: value
            for key, value in module.__dict__.items()
            if key.isupper() or (key.startswith("_") and not key.endswith("_"))
        }
    return {}


def to_dict(C, only_upper_case=True):
    if only_upper_case:
        return get_upper_case_vars_or_protected_of(C)
    else:
        return vars(C)


def parse_arguments(desc, C):
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument(
        "--ENVIRONMENT_NAME",
        "-E",
        type=str,
        default=C.ENVIRONMENT_NAME,
        metavar="ENVIRONMENT_NAME",
        help="Name of the environment to run",
    )
    parser.add_argument(
        "--PRETRAINED_PATH", "-T", metavar="PATH", type=str, default="", help="Path of pre-trained model"
    )
    add_bool_arg(
        parser,
        "render",
        dest="RENDER_ENVIRONMENT",
        default=C.RENDER_ENVIRONMENT,
        help="Rendering of the environment",
    )
    parser.add_argument(
        "--NUM_WORKERS",
        "-N",
        type=int,
        default=4,
        metavar="NUM_WORKERS",
        help="Number of threads for agent (default: 4)",
    )
    add_bool_arg(
        parser,
        "connect_to_running",
        dest="CONNECT_TO_RUNNING",
        default=C.CONNECT_TO_RUNNING,
        help="Connect to already running simulation or start an instance",
    )
    parser.add_argument(
        "--SEED", "-S", type=int, default=C.SEED, metavar="SEED", help=f"Random seed (default: {C.SEED})"
    )
    parser.add_argument(
        "--VERBOSE", "-V", action="store_true", default=C.VERBOSE, help="Enable verbose debug prints"
    )
    parser.add_argument(
        "--skip_confirmation",
        "-skip",
        action="store_true",
        default=False,
        help="Skip confirmation of config to be used",
    )
    parser.add_argument(
        "--ROLLOUTS", "-rollouts", type=int, default=C.ROLLOUTS, metavar="ROLLOUTS", help="Number of rollouts"
    )
    parser.add_argument(
        "--CONFIG",
        "-config",
        type=str,
        default=None,
        metavar="CONFIG",
        help="Path to a config (nullifies all other arguments, if specified)",
    )
    add_bool_arg(parser, "cuda", dest="USE_CUDA", default=C.USE_CUDA, help="Cuda flag")

    args = parser.parse_args()

    return args


def add_bool_arg(parser, name, *, dest=None, default=False, **kwargs):
    if not dest:
        dest = name

    group = parser.add_mutually_exclusive_group(required=False)

    group.add_argument(f"--{name.upper()}", f"-{name.lower()}", dest=dest, action="store_true", **kwargs)
    group.add_argument(
        f"--NO-{name.upper()}", f"-no-{name.lower()}", dest=dest, action="store_false", **kwargs
    )
    parser.set_defaults(**{dest: default})


def check_for_duplicates_in_args(**kwargs) -> None:
    for key, value in kwargs.items():

        occur = 0

        if kwargs.get(key) is not None:
            occur += 1
        else:
            pass

        if key.isupper():
            k_lowered = f"_{key.lower()}"
            if kwargs.get(k_lowered) is not None:
                occur += 1
            else:
                pass
        else:
            k_lowered = f'{key.lstrip("_").upper()}'
            if kwargs.get(k_lowered) is not None:
                occur += 1
            else:
                pass

        if occur > 1:
            warn(f"Config contains hiding duplicates of {key} and {k_lowered}, {occur} times")


def namedtuple_args(n_tuple):
    @wrapt.decorator(adapter=n_tuple)
    def wrapper(wrapped, instance, args, kwargs):
        n = n_tuple(*args, **kwargs)
        return wrapped(n)

    return wrapper



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pytest

from warg.named_ordered_dictionary import NOD, IllegalAttributeKey

__author__ = "cnheider"


def test_attribute_assignment():
    nodict = NOD()
    nodict.paramA = "str_parameter"
    nodict.paramB = 10
    assert nodict.paramA == "str_parameter"
    assert nodict.paramB == 10


def test_dict_arg_construction():
    nodict = NOD({"paramA": "str_parameter", "paramB": 10})
    assert nodict.paramA == "str_parameter"
    assert nodict.paramB == 10


def test_update_and_set_attr_update():
    nodict = NOD()
    nodict.update({"paramA": 20, "paramB": "other_param", "paramC": 5.0})
    nodict.paramA = 10
    assert nodict.paramA == 10
    assert nodict.paramB == "other_param"


def test_set_item_update():
    nodict = NOD()
    nodict["paramA"] = 10
    assert nodict.paramA == 10


def test_as_list_args_and_kwargs():
    vals = [1, 3, 5]
    nodict = NOD(10, val2=2, *vals)
    nolist = nodict.as_list()
    assert nolist[0] == 10
    assert nodict.val2 == 2
    assert nolist[-2] == vals[-1]


def test_kv_unmap():
    nodict = NOD("str_parameter", 10)
    odict = {**nodict}
    assert nodict.as_list()[0] == list(odict.values())[0]


def test_update_attr_with_kwargs():
    nodict = NOD("str_parameter", 10)
    nodict.update(arg1=20, arg0="other_param")
    assert nodict.arg0 == "other_param"
    assert nodict.arg1 == 20


def test_update_attr_with_dict_arg():
    nodict = NOD("str_parameter", 10)
    nodict.update({"arg1": 20, "arg0": "other_param"})
    assert nodict.arg0 == "other_param"
    assert nodict.arg1 == 20


def test_no_existing_attr():
    nodict = NOD(paramA="str_parameter", paramB=10)
    nodict.update(20, "other_param")
    assert nodict.paramB == "other_param"
    assert nodict.paramA == 20
    assert nodict.get("paramC") is None


def test_existing_map_unpacking():
    nodict = NOD(paramA="str_parameter", paramB=10)
    nodict.update(20, "other_param")
    nodict = NOD(paramC="str_parameter", **nodict)
    assert nodict.paramB == "other_param"
    assert nodict.paramA == 20
    assert nodict.get("paramC") is not None


def test_as_list():
    arg0, arg1 = NOD("str_parameter", 10).as_list()
    assert arg0 == "str_parameter"
    assert arg1 == 10


def test_access_operators():
    arg0, arg1 = NOD("str_parameter", 10).as_list()
    columns = NOD.nod_of(arg1, dsa=arg0)
    assert columns["arg1"] == arg1
    assert columns.arg1 == arg1
    assert columns["dsa"] == arg0
    assert columns / "dsa" == arg0
    assert id(columns / "dsa") == id(columns["dsa"])


def test_new_attr_on_existing():
    a = NOD(4, 2)
    a.list = "a"
    assert a.list == "a"


def test_addition_arithmetic():
    a = NOD(4, 2)
    b = a + a
    assert a.as_list() == [8, 4]


def test_illegal_overwrite():
    with pytest.raises(IllegalAttributeKey) as exc_info:
        NOD(update=2)

    assert exc_info.type is IllegalAttributeKey



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import datetime
import os
from warnings import warn

import pkg_resources
from pip._internal.utils.misc import dist_is_editable

__author__ = "cnheider"
__version__ = "0.1.9"
__version_info__ = tuple(int(segment) for segment in __version__.split("."))
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

distributions = {v.key: v for v in pkg_resources.working_set}
distribution = distributions["warg"]
DEVELOP = dist_is_editable(distribution)


def get_version(append_time=DEVELOP):
    version = __version__
    if not version:
        version = os.getenv("VERSION", "0.0.0")

    if append_time:
        now = datetime.datetime.utcnow()
        date_version = now.strftime("%Y%m%d%H%M%S")
        # date_version = time.time()

        if version:
            # Most git tags are prefixed with 'v' (example: v1.2.3) this is
            # never desirable for artifact repositories, so we strip the
            # leading 'v' if it's present.
            version = version[1:] if isinstance(version, str) and version.startswith("v") else version
        else:
            # Default version is an ISO8601 compliant datetime. PyPI doesn't allow
            # the colon ':' character in its versions, and time is required to allow
            # for multiple publications to master in one day. This datetime string
            # uses the 'basic' ISO8601 format for both its date and time components
            # to avoid issues with the colon character (ISO requires that date and
            # time components of a date-time string must be uniformly basic or
            # extended, which is why the date component does not have dashes.
            #
            # Publications using datetime versions should only be made from master
            # to represent the HEAD moving forward.
            warn(f"Environment variable VERSION is not set, only using datetime: {date_version}")

            # warn(f'Environment variable VERSION is not set, only using timestamp: {version}')

        version = f"{version}.{date_version}"

    return version


if __version__ is None:
    __version__ = get_version(append_time=True)

__version_info__ = tuple(int(segment) for segment in __version__.split("."))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


def python_version_check():
    import sys

    assert sys.version_info.major == 3 and sys.version_info.minor >= 6, (
        f"This project is utilises language features only present Python 3.6 and greater. "
        f"You are running {sys.version_info}."
    )


python_version_check()

from WargPackage import WargPackage

__author__ = "cnheider"

from setuptools import setup

if __name__ == "__main__":

    pkg = WargPackage()

    setup(
        name=pkg.package_name,
        version=pkg.version,
        packages=pkg.packages,
        package_data=pkg.package_data,
        author=pkg.author_name,
        author_email=pkg.author_email,
        maintainer=pkg.maintainer_name,
        maintainer_email=pkg.maintainer_email,
        description=pkg.description,
        license=pkg.license,
        keywords=pkg.keyword,
        url=pkg.url,
        download_url=pkg.download_url,
        install_requires=pkg.requirements,
        extras_require=pkg.extras,
        setup_requires=pkg.setup_dependencies,
        entry_points=pkg.entry_points,
        classifiers=pkg.classifiers,
        long_description_content_type=pkg.readme_type,
        long_description=pkg.readme,
        tests_require=pkg.test_dependencies,
        include_package_data=True,
        python_requires=">=3",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import gi

gi.require_version("Gtk", "3.0")
from gi.repository import GdkPixbuf

__author__ = "cnheider"
__doc__ = "Based on the notifications spec at: http://developer.gnome.org/notification-spec/"
__version__ = "0.0.1"

import time

import dbus

from warg.utilities import action_callback, closed_callback, no_op

EXPIRES_DEFAULT = -1
EXPIRES_NEVER = 0

URGENCY_LOW = 0
URGENCY_NORMAL = 1
URGENCY_CRITICAL = 2
urgency_levels = [URGENCY_LOW, URGENCY_NORMAL, URGENCY_CRITICAL]

IS_SETUP = False
APP_NAME = f"unnamed_app_{time.time()}"
HAVE_MAINLOOP = False

NOTIFICATIONS_REGISTRY = {}


# TODO: Object orient globals!


class NotSetupError(RuntimeError):
    """Error raised if you try to communicate with the server before calling
:func:`init`.
"""

    pass


class UnconstructedDbusObject(object):
    def __getattr__(self, name):
        raise NotSetupError("You must call toaster.init() first")


dbus_interface = UnconstructedDbusObject()


def init(app_name, mainloop=None):
    """Initialise the D-Bus connection. Must be called before you send any
notifications, or retrieve server info or capabilities.

To get callbacks from notifications, DBus must be integrated with a mainloop.
There are three ways to achieve this:

- Set a default mainloop (dbus.set_default_main_loop) before calling init()
- Pass the mainloop parameter as a string 'glib' or 'qt' to integrate with
  those mainloops. (N.B. passing 'qt' currently makes that the default dbus
  mainloop, because that's the only way it seems to work.)
- Pass the mainloop parameter a DBus compatible mainloop instance, such as
  dbus.mainloop.glib.DBusGMainLoop().

If you only want to display notifications, without receiving information
back from them, you can safely omit mainloop.
"""
    global APP_NAME, IS_SETUP, dbus_interface, HAVE_MAINLOOP

    if mainloop == "glib":
        from dbus.mainloop.glib import DBusGMainLoop

        mainloop = DBusGMainLoop()
    elif mainloop == "qt":
        from dbus.mainloop.qt import DBusQtMainLoop

        # For some reason, this only works if we make it the default mainloop
        # for dbus. That might make life tricky for anyone trying to juggle two
        # event loops, but I can't see any way round it.
        mainloop = DBusQtMainLoop(set_as_default=True)

    bus = dbus.SessionBus(mainloop=mainloop)

    dbus_obj = bus.get_object("org.freedesktop.Notifications", "/org/freedesktop/Notifications")
    dbus_interface = dbus.Interface(dbus_obj, dbus_interface="org.freedesktop.Notifications")
    APP_NAME = app_name
    IS_SETUP = True

    if mainloop or dbus.get_default_main_loop():
        HAVE_MAINLOOP = True
        dbus_interface.connect_to_signal("ActionInvoked", action_callback)
        dbus_interface.connect_to_signal("NotificationClosed", closed_callback)

    return True


def is_initted():
    """Has init() been called? Only exists for compatibility with pynotify.
"""
    return IS_SETUP


def get_app_name():
    """Return appname. Only exists for compatibility with pynotify.
"""
    return APP_NAME


def de_init():
    """Undo what init() does."""
    global IS_SETUP, dbus_interface, HAVE_MAINLOOP
    IS_SETUP = False
    HAVE_MAINLOOP = False
    dbus_interface = UnconstructedDbusObject()


# Retrieve basic server information --------------------------------------------


def get_server_caps():
    """Get a list of server capabilities.

These are short strings, listed `in the spec
<http://people.gnome.org/~mccann/docs/notification-spec/notification-spec-latest.html#commands>`_.
Vendors may also list extra capabilities with an 'x-' prefix, e.g. 'x-canonical-append'.
"""
    return [str(x) for x in dbus_interface.GetCapabilities()]


def get_server_info():
    """Get basic information about the server.
"""
    res = dbus_interface.GetServerInformation()
    return {"name": str(res[0]), "vendor": str(res[1]), "version": str(res[2]), "spec-version": str(res[3])}


class Toast(object):
    """A notification object.

summary : str
  The title text
message : str
  The body text, if the server has the 'body' capability.
icon : str
  Path to an icon image, or the name of a stock icon. Stock icons available
  in Ubuntu are `listed here <https://wiki.ubuntu.com/NotificationDevelopmentGuidelines
  #How_do_I_get_these_slick_icons>`_.
  You can also set an icon from data in your application - see
  :meth:`set_icon_from_pixbuf`.
"""

    _id = 0
    _timeout = -1  # -1 = server default settings
    _closed_callback = no_op

    def __init__(self, title, body="", *, icon=""):
        self.title = title
        self.body = body
        self._hints = {}

        if isinstance(icon, GdkPixbuf.Pixbuf):
            self._icon = ""
            self.set_hint("icon_data", icon)
        else:
            self._icon = icon

        self._actions = {}
        self._data = {}  # Any data the user wants to attach

    def show(self):
        """Ask the server to show the notification.

Call this after you have finished setting any parameters of the
notification that you want.
"""
        nid = dbus_interface.Notify(
            APP_NAME,  # app_name       (spec names)
            self._id,  # replaces_id
            self._icon,  # app_icon
            self.title,  # summary
            self.body,  # body
            self._make_actions_array(),  # actions
            self._hints,  # hints
            self._timeout,  # expire_timeout
        )

        self._id = int(nid)

        if HAVE_MAINLOOP:
            NOTIFICATIONS_REGISTRY[self._id] = self
        return True

    def update(self, title, body="", *, icon=None):
        """Replace the summary and body of the notification, and optionally its
icon. You should call :meth:`show` again after this to display the
updated notification.
"""
        self.title = title
        self.body = body
        if icon is not None:
            self._icon = icon

    def close(self):
        """Ask the server to close this notification."""
        if self._id != 0:
            dbus_interface.CloseNotification(self._id)

    def set_hint(self, key, value):
        """n.set_hint(key, value) <--> n.hints[key] = value

See `hints in the spec <http://people.gnome.org/~mccann/docs/notification-spec/notification-spec-latest
.html#hints>`_.

Only exists for compatibility with pynotify.
"""
        self._hints[key] = value

    set_hint_string = set_hint_int32 = set_hint_double = set_hint

    def set_hint_byte(self, key, value):
        """Set a hint with a dbus byte value. The input value can be an
integer or a bytes string of length 1.
"""
        self._hints[key] = dbus.Byte(value)

    def set_urgency(self, level):
        """Set the urgency level to one of URGENCY_LOW, URGENCY_NORMAL or
URGENCY_CRITICAL.
"""
        if level not in urgency_levels:
            raise ValueError("Unknown urgency level specified", level)
        self.set_hint_byte("urgency", level)

    def set_category(self, category):
        """Set the 'category' hint for this notification.

See `categories in the spec <http://people.gnome.org/~mccann/docs/notification-spec/notification-spec
-latest.html#categories>`_.
"""
        self._hints["category"] = category

    def set_timeout(self, timeout):
        """Set the display duration in milliseconds, or one of the special
values EXPIRES_DEFAULT or EXPIRES_NEVER. This is a request, which the
server might ignore.

Only exists for compatibility with pynotify; you can simply set::

  n.timeout = 5000
"""
        if not isinstance(timeout, int):
            raise TypeError("timeout value was not int", timeout)
        self._timeout = timeout

    def get_timeout(self):
        """Return the timeout value for this notification.

Only exists for compatibility with pynotify; you can inspect the
timeout attribute directly.
"""
        return self._timeout

    def add_action(self, action, label, callback, user_data=None):
        """Add an action to the notification.

Check for the 'actions' server capability before using this.

action : str
  A brief key.
label : str
  The text displayed on the action button
callback : callable
  A function taking at 2-3 parameters: the Notification object, the
  action key and (if specified) the user_data.
user_data :
  An extra argument to pass to the callback.
"""
        self._actions[action] = (label, callback, user_data)

    def _make_actions_array(self):
        """Make the actions array to send over DBus.
"""
        arr = []
        for action, (label, callback, user_data) in self._actions.items():
            arr.append(action)
            arr.append(label)
        return arr

    def _action_callback(self, action):
        """Called when the user selects an action on the notification, to
dispatch it to the relevant user-specified callback.
"""
        try:
            label, callback, user_data = self._actions[action]
        except KeyError:
            return

        if user_data is None:
            callback(self, action)
        else:
            callback(self, action, user_data)

    def connect(self, event, callback):
        """Set the callback for the notification closing; the only valid value
for event is 'closed' (the parameter is kept for compatibility with pynotify).

The callback will be called with the :class:`Notification` instance.
"""
        if event != "closed":
            raise ValueError("'closed' is the only valid value for event", event)
        self._closed_callback = callback

    def set_data(self, key, value):
        """n.set_data(key, value) <--> n.data[key] = value

Only exists for compatibility with pynotify.
"""
        self._data[key] = value

    def get_data(self, key):
        """n.get_data(key) <--> n.data[key]

Only exists for compatibility with pynotify.
"""
        return self._data[key]

    def set_icon_from_pixbuf(self, icon):
        """Set a custom icon from a GdkPixbuf.
"""
        self._hints["icon_data"] = self._get_icon_struct(icon)

    @staticmethod
    def _get_icon_struct(icon):
        return (
            icon.get_width(),
            icon.get_height(),
            icon.get_rowstride(),
            icon.get_has_alpha(),
            icon.get_bits_per_sample(),
            icon.get_n_channels(),
            dbus.ByteArray(icon.get_pixels()),
        )

    def set_location(self, x, y):
        """Set the notification location as (x, y), if the server supports it.
"""
        if (not isinstance(x, int)) or (not isinstance(y, int)):
            raise TypeError("x and y must both be ints", (x, y))
        self._hints["x"] = x
        self._hints["y"] = y


if __name__ == "__main__":
    import gi

    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk

    init("Test")

    helper = Gtk.Button()
    a_icon = helper.render_icon(Gtk.STOCK_DIALOG_INFO, Gtk.IconSize.DIALOG)

    t = Toast("Title", "Body")
    t.set_icon_from_pixbuf(a_icon)
    for i in range(10):
        t.title = f"Title{i}"
        t.body = f"Body{i}"
        t.show()
        time.sleep(0.1)
        if i == 4:
            a_icon = helper.render_icon(Gtk.STOCK_DIALOG_QUESTION, Gtk.IconSize.DIALOG)
            t.set_icon_from_pixbuf(a_icon)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "cnheider"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""


class SingletonBase:
    instance = None

    def __new__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__new__(cls, *args, **kwargs)

        return cls.instance


class SingletonMeta(type):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        cls.instance = None

    def __call__(cls, *args, **kwargs):
        if cls.instance is None:
            cls.instance = super().__call__(*args, **kwargs)

        return cls.instance


class SingletonBaseClass(SingletonBase):
    pass


class S1(SingletonBaseClass):
    pass


class SingletonBaseMeta(metaclass=SingletonMeta):
    pass


class S2(SingletonBaseMeta):
    pass


if __name__ == "__main__":
    print(SingletonBaseClass())
    print(SingletonBaseClass())
    print(S1())

    print(SingletonBaseMeta())
    print(SingletonBaseMeta())
    print(S2())
